# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/logic.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1frelationalai/lqp/v1/logic.proto\x12\x13relationalai.lqp.v1\"\x80\x01\n\x0b\x44\x65\x63laration\x12,\n\x03\x64\x65\x66\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.DefH\x00R\x03\x64\x65\x66\x12/\n\x04loop\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.LoopH\x00R\x04loopB\x12\n\x10\x64\x65\x63laration_type\"\xa6\x01\n\x03\x44\x65\x66\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xaf\x01\n\x04Loop\x12\x41\n\x0ctemporal_var\x18\x01 \x01(\x0b\x32\x1e.relationalai.lqp.v1.LoopIndexR\x0btemporalVar\x12.\n\x05inits\x18\x02 \x03(\x0b\x32\x18.relationalai.lqp.v1.DefR\x05inits\x12\x34\n\x04\x62ody\x18\x03 \x03(\x0b\x32 .relationalai.lqp.v1.DeclarationR\x04\x62ody\"o\n\x0b\x41\x62straction\x12,\n\x04vars\x18\x01 \x03(\x0b\x32\x18.relationalai.lqp.v1.VarR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\x83\x05\n\x07\x46ormula\x12\x35\n\x06\x65xists\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ExistsH\x00R\x06\x65xists\x12\x35\n\x06reduce\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ReduceH\x00R\x06reduce\x12\x44\n\x0b\x63onjunction\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.ConjunctionH\x00R\x0b\x63onjunction\x12\x44\n\x0b\x64isjunction\x18\x04 \x01(\x0b\x32 .relationalai.lqp.v1.DisjunctionH\x00R\x0b\x64isjunction\x12,\n\x03not\x18\x05 \x01(\x0b\x32\x18.relationalai.lqp.v1.NotH\x00R\x03not\x12,\n\x03\x66\x66i\x18\x06 \x01(\x0b\x32\x18.relationalai.lqp.v1.FFIH\x00R\x03\x66\x66i\x12/\n\x04\x61tom\x18\x07 \x01(\x0b\x32\x19.relationalai.lqp.v1.AtomH\x00R\x04\x61tom\x12\x35\n\x06pragma\x18\x08 \x01(\x0b\x32\x1b.relationalai.lqp.v1.PragmaH\x00R\x06pragma\x12>\n\tprimitive\x18\t \x01(\x0b\x32\x1e.relationalai.lqp.v1.PrimitiveH\x00R\tprimitive\x12\x39\n\x08rel_atom\x18\n \x01(\x0b\x32\x1c.relationalai.lqp.v1.RelAtomH\x00R\x07relAtom\x12/\n\x04\x63\x61st\x18\x0b \x01(\x0b\x32\x19.relationalai.lqp.v1.CastH\x00R\x04\x63\x61stB\x0e\n\x0c\x66ormula_type\">\n\x06\x45xists\x12\x34\n\x04\x62ody\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\"\xa1\x01\n\x06Reduce\x12\x30\n\x02op\x18\x01 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x02op\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"?\n\x0b\x43onjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"?\n\x0b\x44isjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"5\n\x03Not\x12.\n\x03\x61rg\x18\x01 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x03\x61rg\"\x80\x01\n\x03\x46\x46I\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x34\n\x04\x61rgs\x18\x02 \x03(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x61rgs\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"l\n\x04\x41tom\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"M\n\x06Pragma\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"S\n\tPrimitive\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x32\n\x05terms\x18\x02 \x03(\x0b\x32\x1c.relationalai.lqp.v1.RelTermR\x05terms\"Q\n\x07RelAtom\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\x12\x32\n\x05terms\x18\x02 \x03(\x0b\x32\x1c.relationalai.lqp.v1.RelTermR\x05terms\"\x9c\x01\n\x04\x43\x61st\x12\x30\n\x04type\x18\x01 \x01(\x0b\x32\x1c.relationalai.lqp.v1.RelTypeR\x04type\x12/\n\x05input\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05input\x12\x31\n\x06result\x18\x03 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermR\x06result\"\xa1\x01\n\x07RelTerm\x12T\n\x11specialized_value\x18\x01 \x01(\x0b\x32%.relationalai.lqp.v1.SpecializedValueH\x00R\x10specializedValue\x12/\n\x04term\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermH\x00R\x04termB\x0f\n\rrel_term_type\"~\n\x04Term\x12,\n\x03var\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.VarH\x00R\x03var\x12;\n\x08\x63onstant\x18\x02 \x01(\x0b\x32\x1d.relationalai.lqp.v1.ConstantH\x00R\x08\x63onstantB\x0b\n\tterm_type\"K\n\x03Var\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x30\n\x04type\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.RelTypeR\x04type\"E\n\x08\x43onstant\x12\x39\n\x05value\x18\x01 \x01(\x0b\x32#.relationalai.lqp.v1.PrimitiveValueR\x05value\"M\n\x10SpecializedValue\x12\x39\n\x05value\x18\x01 \x01(\x0b\x32#.relationalai.lqp.v1.PrimitiveValueR\x05value\"R\n\tAttribute\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x31\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x1d.relationalai.lqp.v1.ConstantR\x04\x61rgs\"<\n\nRelationId\x12\x15\n\x06id_low\x18\x01 \x01(\x06R\x05idLow\x12\x17\n\x07id_high\x18\x02 \x01(\x06R\x06idHigh\"\x1f\n\tLoopIndex\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\xa6\x01\n\x07RelType\x12K\n\x0eprimitive_type\x18\x01 \x01(\x0e\x32\".relationalai.lqp.v1.PrimitiveTypeH\x00R\rprimitiveType\x12\x42\n\nvalue_type\x18\x02 \x01(\x0e\x32!.relationalai.lqp.v1.RelValueTypeH\x00R\tvalueTypeB\n\n\x08rel_type\"\xc5\x01\n\x0ePrimitiveValue\x12#\n\x0cstring_value\x18\x01 \x01(\tH\x00R\x0bstringValue\x12\x1d\n\tint_value\x18\x02 \x01(\x03H\x00R\x08intValue\x12!\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00R\nfloatValue\x12\x43\n\ruint128_value\x18\x04 \x01(\x0b\x32\x1c.relationalai.lqp.v1.UInt128H\x00R\x0cuint128ValueB\x07\n\x05value\"/\n\x07UInt128\x12\x10\n\x03low\x18\x01 \x01(\x06R\x03low\x12\x12\n\x04high\x18\x02 \x01(\x06R\x04high*\x98\x01\n\rPrimitiveType\x12\x1e\n\x1aPRIMITIVE_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15PRIMITIVE_TYPE_STRING\x10\x01\x12\x16\n\x12PRIMITIVE_TYPE_INT\x10\x02\x12\x18\n\x14PRIMITIVE_TYPE_FLOAT\x10\x03\x12\x1a\n\x16PRIMITIVE_TYPE_UINT128\x10\x04*\x92\x03\n\x0cRelValueType\x12\x1e\n\x1aREL_VALUE_TYPE_UNSPECIFIED\x10\x00\x12\x1a\n\x16REL_VALUE_TYPE_DECIMAL\x10\x01\x12\x17\n\x13REL_VALUE_TYPE_DATE\x10\x02\x12\x1b\n\x17REL_VALUE_TYPE_DATETIME\x10\x03\x12\x1d\n\x19REL_VALUE_TYPE_NANOSECOND\x10\x04\x12\x1e\n\x1aREL_VALUE_TYPE_MICROSECOND\x10\x05\x12\x1e\n\x1aREL_VALUE_TYPE_MILLISECOND\x10\x06\x12\x19\n\x15REL_VALUE_TYPE_SECOND\x10\x07\x12\x19\n\x15REL_VALUE_TYPE_MINUTE\x10\x08\x12\x17\n\x13REL_VALUE_TYPE_HOUR\x10\t\x12\x16\n\x12REL_VALUE_TYPE_DAY\x10\n\x12\x17\n\x13REL_VALUE_TYPE_WEEK\x10\x0b\x12\x18\n\x14REL_VALUE_TYPE_MONTH\x10\x0c\x12\x17\n\x13REL_VALUE_TYPE_YEAR\x10\rb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.logic_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PRIMITIVETYPE']._serialized_start=3470
  _globals['_PRIMITIVETYPE']._serialized_end=3622
  _globals['_RELVALUETYPE']._serialized_start=3625
  _globals['_RELVALUETYPE']._serialized_end=4027
  _globals['_DECLARATION']._serialized_start=57
  _globals['_DECLARATION']._serialized_end=185
  _globals['_DEF']._serialized_start=188
  _globals['_DEF']._serialized_end=354
  _globals['_LOOP']._serialized_start=357
  _globals['_LOOP']._serialized_end=532
  _globals['_ABSTRACTION']._serialized_start=534
  _globals['_ABSTRACTION']._serialized_end=645
  _globals['_FORMULA']._serialized_start=648
  _globals['_FORMULA']._serialized_end=1291
  _globals['_EXISTS']._serialized_start=1293
  _globals['_EXISTS']._serialized_end=1355
  _globals['_REDUCE']._serialized_start=1358
  _globals['_REDUCE']._serialized_end=1519
  _globals['_CONJUNCTION']._serialized_start=1521
  _globals['_CONJUNCTION']._serialized_end=1584
  _globals['_DISJUNCTION']._serialized_start=1586
  _globals['_DISJUNCTION']._serialized_end=1649
  _globals['_NOT']._serialized_start=1651
  _globals['_NOT']._serialized_end=1704
  _globals['_FFI']._serialized_start=1707
  _globals['_FFI']._serialized_end=1835
  _globals['_ATOM']._serialized_start=1837
  _globals['_ATOM']._serialized_end=1945
  _globals['_PRAGMA']._serialized_start=1947
  _globals['_PRAGMA']._serialized_end=2024
  _globals['_PRIMITIVE']._serialized_start=2026
  _globals['_PRIMITIVE']._serialized_end=2109
  _globals['_RELATOM']._serialized_start=2111
  _globals['_RELATOM']._serialized_end=2192
  _globals['_CAST']._serialized_start=2195
  _globals['_CAST']._serialized_end=2351
  _globals['_RELTERM']._serialized_start=2354
  _globals['_RELTERM']._serialized_end=2515
  _globals['_TERM']._serialized_start=2517
  _globals['_TERM']._serialized_end=2643
  _globals['_VAR']._serialized_start=2645
  _globals['_VAR']._serialized_end=2720
  _globals['_CONSTANT']._serialized_start=2722
  _globals['_CONSTANT']._serialized_end=2791
  _globals['_SPECIALIZEDVALUE']._serialized_start=2793
  _globals['_SPECIALIZEDVALUE']._serialized_end=2870
  _globals['_ATTRIBUTE']._serialized_start=2872
  _globals['_ATTRIBUTE']._serialized_end=2954
  _globals['_RELATIONID']._serialized_start=2956
  _globals['_RELATIONID']._serialized_end=3016
  _globals['_LOOPINDEX']._serialized_start=3018
  _globals['_LOOPINDEX']._serialized_end=3049
  _globals['_RELTYPE']._serialized_start=3052
  _globals['_RELTYPE']._serialized_end=3218
  _globals['_PRIMITIVEVALUE']._serialized_start=3221
  _globals['_PRIMITIVEVALUE']._serialized_end=3418
  _globals['_UINT128']._serialized_start=3420
  _globals['_UINT128']._serialized_end=3467
# @@protoc_insertion_point(module_scope)



import lqp.ir as ir
from typing import Union, Dict, Any, List, Tuple, Set, cast

class ValidationError(Exception):
    pass

class LqpVisitor:
    def visit(self, node: ir.LqpNode, *args: Any) -> None:
        method_name = f'visit_{node.__class__.__name__}'
        visitor_method = getattr(self, method_name, self.generic_visit)
        return visitor_method(node, *args)

    def generic_visit(self, node: ir.LqpNode, *args: Any) -> None:
        for name, _ in node.__dataclass_fields__.items():
            value = getattr(node, name)
            if isinstance(value, ir.LqpNode):
                self.visit(value, *args)
            elif isinstance(value, (list, tuple)):
                for item in value:
                    if isinstance(item, ir.LqpNode):
                        self.visit(item, *args)
            elif isinstance(value, dict):
                 for key, item in value.items():
                    if isinstance(item, ir.LqpNode):
                        self.visit(item, *args)

class VariableScopeVisitor(LqpVisitor):
    def __init__(self, fill_types=False):
        self.scopes: list[Dict[str, ir.RelType]] = []
        self.fill_types = fill_types

    def _declare_var(self, var: ir.Var):
        if var.name in self.scopes[-1]:
            pass # Allow shadowing
        self.scopes[-1][var.name] = var.type

    def _get_type_name(self, rel_type: ir.RelType) -> str:
        if isinstance(rel_type, (ir.PrimitiveType, ir.RelValueType)):
            return rel_type.name
        return "UNKNOWN"

    def _get_type(self, var: str) -> Union[ir.RelType, None]:
        for scope in reversed(self.scopes):
            if var in scope:
                return scope[var]

    def _check_var_usage(self, var: ir.Var):
        declared_type: Union[ir.RelType, None] = self._get_type(var.name)
        if declared_type is None and not self.fill_types:
            raise ValidationError(f"Undeclared variable used: '{var.name}'")
        if var.type == ir.PrimitiveType.UNSPECIFIED and self.fill_types:
            object.__setattr__(var, 'type', declared_type)
        elif var.type != declared_type and not self.fill_types:
            type_name_declared = self._get_type_name(cast(ir.RelType, declared_type))
            type_name_used = self._get_type_name(var.type)
            raise ValidationError(
                f"Type mismatch for variable '{var.name}': "
                f"Declared as {type_name_declared}, used as {type_name_used}"
            )

    def visit_Abstraction(self, node: ir.Abstraction):
        self.scopes.append({})
        for var in node.vars:
            self._declare_var(var)
        self.visit(node.value)
        self.scopes.pop()

    def visit_Var(self, node: ir.Var):
        self._check_var_usage(node)

class UnusedVariableVisitor(LqpVisitor):
    def __init__(self):
        self.scopes: List[Tuple[Set[str], Set[str]]] = []

    def _declare_var(self, var_name: str):
        if self.scopes:
            self.scopes[-1][0].add(var_name)

    def _mark_var_used(self, var_name: str):
        for declared, used in reversed(self.scopes):
            if var_name in declared:
                used.add(var_name)
                break

    def visit_Abstraction(self, node: ir.Abstraction):
        self.scopes.append((set(), set()))
        for var in node.vars:
            self._declare_var(var.name)
        self.visit(node.value)
        declared, used = self.scopes.pop()
        unused = declared - used
        if unused:
            for var_name in unused:
                raise ValidationError(f"Unused variable declared: '{var_name}'")

    def visit_Var(self, node: ir.Var, *args: Any):
        self._mark_var_used(node.name)

def validate_lqp(lqp: ir.LqpNode):
    VariableScopeVisitor(fill_types=False).visit(lqp)
    UnusedVariableVisitor().visit(lqp)

def fill_types(lqp: ir.LqpNode):
    VariableScopeVisitor(fill_types=True).visit(lqp)



import os
import re
import pytest
import sys
from pathlib import Path
from lqp.parser import parse_lqp
from lqp.emit import ir_to_proto
from lqp.validator import ValidationError, validate_lqp

TEST_INPUTS_DIR = Path(__file__).parent / "test_files" / "lqp_input"
TEST_OUTPUTS_DIR = Path(__file__).parent / "test_files" / "bin_output"

def get_all_input_files():
    """Find all .lqp files in the test inputs directory and subdirectories"""
    input_files = []

    for root, dirs, files in os.walk(TEST_INPUTS_DIR):
        for file in files:
            if file.endswith(".lqp"):
                input_files.append(os.path.join(root, file))

    return input_files

def get_output_file(input_file):
    """Get the corresponding output file for a given input file"""
    base_name = os.path.basename(input_file)
    output_file = os.path.join(TEST_OUTPUTS_DIR, base_name.replace(".lqp", ".bin"))
    return output_file

@pytest.mark.parametrize("input_file", get_all_input_files())
def test_parse_lqp(input_file):
    """Test that each input file can be successfully parsed"""
    try:
        with open(input_file, "r") as f:
            content = f.read()

        # Parse the file and check it returns a valid protobuf object
        result = ir_to_proto(parse_lqp(content))
        assert result is not None, f"Failed to parse {input_file}"

        # Log the successful parse for verbose output
        print(f"Successfully parsed {input_file}")

        # Check that the generated proto binary matches the expected output
        output_file = get_output_file(input_file)
        with open(output_file, "rb") as f:
            expected_output = f.read()
            assert result.SerializeToString() == expected_output, f"Output does not match for {input_file}"

    except Exception as e:
        pytest.fail(f"Failed checking {input_file}: {str(e)}")

VALIDATOR_DIR = Path(__file__).parent / "validator"

def test_valid_validator_files():
    for validator_file in VALIDATOR_DIR.glob("valid_*.lqp"):
        with open(validator_file, "r") as f:
            content = f.read()
        try:
            result = parse_lqp(content)
            assert result is not None, f"Failed to parse {validator_file}"
            print(f"Successfully validated {validator_file}")
        except Exception as e:
            pytest.fail(f"Failed to parse valid validator file {validator_file}: {str(e)}")

def extract_expected_error(file_path):
    with open(file_path, "r") as f:
        content = f.read()
    error_match = re.search(r';;\s*ERROR:\s*(.+)(?:\n|\r\n?)', content)
    if error_match:
        return error_match.group(1).strip()
    return None

@pytest.mark.parametrize("validator_file", [f for f in os.listdir(VALIDATOR_DIR) if f.startswith("fail_")])
def test_validator_failure_files(validator_file):
    file_path = VALIDATOR_DIR / validator_file
    expected_error = extract_expected_error(file_path)
    if not expected_error:
        pytest.skip(f"No expected error comment found in {validator_file}")
        return
    with open(file_path, "r") as f:
        content = f.read()
    result = parse_lqp(content)
    with pytest.raises(ValidationError) as exc_info:
        validate_lqp(result)
    error_message = str(exc_info.value)
    assert expected_error in error_message, f"Expected '{expected_error}' in error message: {error_message}"



from lqp.proto.v1 import logic_pb2 as _logic_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Fragment(_message.Message):
    __slots__ = ("id", "declarations")
    ID_FIELD_NUMBER: _ClassVar[int]
    DECLARATIONS_FIELD_NUMBER: _ClassVar[int]
    id: FragmentId
    declarations: _containers.RepeatedCompositeFieldContainer[_logic_pb2.Declaration]
    def __init__(self, id: _Optional[_Union[FragmentId, _Mapping]] = ..., declarations: _Optional[_Iterable[_Union[_logic_pb2.Declaration, _Mapping]]] = ...) -> None: ...

class FragmentId(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: bytes
    def __init__(self, id: _Optional[bytes] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/transactions.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from lqp.proto.v1 import fragments_pb2 as relationalai_dot_lqp_dot_v1_dot_fragments__pb2
from lqp.proto.v1 import logic_pb2 as relationalai_dot_lqp_dot_v1_dot_logic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&relationalai/lqp/v1/transactions.proto\x12\x13relationalai.lqp.v1\x1a#relationalai/lqp/v1/fragments.proto\x1a\x1frelationalai/lqp/v1/logic.proto\"A\n\x0bTransaction\x12\x32\n\x06\x65pochs\x18\x01 \x03(\x0b\x32\x1a.relationalai.lqp.v1.EpochR\x06\x65pochs\"\xc0\x01\n\x05\x45poch\x12G\n\x11persistent_writes\x18\x01 \x03(\x0b\x32\x1a.relationalai.lqp.v1.WriteR\x10persistentWrites\x12=\n\x0clocal_writes\x18\x02 \x03(\x0b\x32\x1a.relationalai.lqp.v1.WriteR\x0blocalWrites\x12/\n\x05reads\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.ReadR\x05reads\"\xc3\x01\n\x05Write\x12\x35\n\x06\x64\x65\x66ine\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.DefineH\x00R\x06\x64\x65\x66ine\x12;\n\x08undefine\x18\x02 \x01(\x0b\x32\x1d.relationalai.lqp.v1.UndefineH\x00R\x08undefine\x12\x38\n\x07\x63ontext\x18\x03 \x01(\x0b\x32\x1c.relationalai.lqp.v1.ContextH\x00R\x07\x63ontextB\x0c\n\nwrite_type\"C\n\x06\x44\x65\x66ine\x12\x39\n\x08\x66ragment\x18\x01 \x01(\x0b\x32\x1d.relationalai.lqp.v1.FragmentR\x08\x66ragment\"L\n\x08Undefine\x12@\n\x0b\x66ragment_id\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.FragmentIdR\nfragmentId\"H\n\x07\x43ontext\x12=\n\trelations\x18\x01 \x03(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\trelations\"\xed\x01\n\x04Read\x12\x35\n\x06\x64\x65mand\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.DemandH\x00R\x06\x64\x65mand\x12\x35\n\x06output\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.OutputH\x00R\x06output\x12\x36\n\x07what_if\x18\x03 \x01(\x0b\x32\x1b.relationalai.lqp.v1.WhatIfH\x00R\x06whatIf\x12\x32\n\x05\x61\x62ort\x18\x04 \x01(\x0b\x32\x1a.relationalai.lqp.v1.AbortH\x00R\x05\x61\x62ortB\x0b\n\tread_type\"J\n\x06\x44\x65mand\x12@\n\x0brelation_id\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationId\"^\n\x06Output\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12@\n\x0brelation_id\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationId\"R\n\x06WhatIf\x12\x16\n\x06\x62ranch\x18\x01 \x01(\tR\x06\x62ranch\x12\x30\n\x05\x65poch\x18\x02 \x01(\x0b\x32\x1a.relationalai.lqp.v1.EpochR\x05\x65poch\"]\n\x05\x41\x62ort\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12@\n\x0brelation_id\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationIdb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.transactions_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_TRANSACTION']._serialized_start=133
  _globals['_TRANSACTION']._serialized_end=198
  _globals['_EPOCH']._serialized_start=201
  _globals['_EPOCH']._serialized_end=393
  _globals['_WRITE']._serialized_start=396
  _globals['_WRITE']._serialized_end=591
  _globals['_DEFINE']._serialized_start=593
  _globals['_DEFINE']._serialized_end=660
  _globals['_UNDEFINE']._serialized_start=662
  _globals['_UNDEFINE']._serialized_end=738
  _globals['_CONTEXT']._serialized_start=740
  _globals['_CONTEXT']._serialized_end=812
  _globals['_READ']._serialized_start=815
  _globals['_READ']._serialized_end=1052
  _globals['_DEMAND']._serialized_start=1054
  _globals['_DEMAND']._serialized_end=1128
  _globals['_OUTPUT']._serialized_start=1130
  _globals['_OUTPUT']._serialized_end=1224
  _globals['_WHATIF']._serialized_start=1226
  _globals['_WHATIF']._serialized_end=1308
  _globals['_ABORT']._serialized_start=1310
  _globals['_ABORT']._serialized_end=1403
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PrimitiveType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PRIMITIVE_TYPE_UNSPECIFIED: _ClassVar[PrimitiveType]
    PRIMITIVE_TYPE_STRING: _ClassVar[PrimitiveType]
    PRIMITIVE_TYPE_INT: _ClassVar[PrimitiveType]
    PRIMITIVE_TYPE_FLOAT: _ClassVar[PrimitiveType]
    PRIMITIVE_TYPE_UINT128: _ClassVar[PrimitiveType]

class RelValueType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    REL_VALUE_TYPE_UNSPECIFIED: _ClassVar[RelValueType]
    REL_VALUE_TYPE_DECIMAL: _ClassVar[RelValueType]
    REL_VALUE_TYPE_DATE: _ClassVar[RelValueType]
    REL_VALUE_TYPE_DATETIME: _ClassVar[RelValueType]
    REL_VALUE_TYPE_NANOSECOND: _ClassVar[RelValueType]
    REL_VALUE_TYPE_MICROSECOND: _ClassVar[RelValueType]
    REL_VALUE_TYPE_MILLISECOND: _ClassVar[RelValueType]
    REL_VALUE_TYPE_SECOND: _ClassVar[RelValueType]
    REL_VALUE_TYPE_MINUTE: _ClassVar[RelValueType]
    REL_VALUE_TYPE_HOUR: _ClassVar[RelValueType]
    REL_VALUE_TYPE_DAY: _ClassVar[RelValueType]
    REL_VALUE_TYPE_WEEK: _ClassVar[RelValueType]
    REL_VALUE_TYPE_MONTH: _ClassVar[RelValueType]
    REL_VALUE_TYPE_YEAR: _ClassVar[RelValueType]
PRIMITIVE_TYPE_UNSPECIFIED: PrimitiveType
PRIMITIVE_TYPE_STRING: PrimitiveType
PRIMITIVE_TYPE_INT: PrimitiveType
PRIMITIVE_TYPE_FLOAT: PrimitiveType
PRIMITIVE_TYPE_UINT128: PrimitiveType
REL_VALUE_TYPE_UNSPECIFIED: RelValueType
REL_VALUE_TYPE_DECIMAL: RelValueType
REL_VALUE_TYPE_DATE: RelValueType
REL_VALUE_TYPE_DATETIME: RelValueType
REL_VALUE_TYPE_NANOSECOND: RelValueType
REL_VALUE_TYPE_MICROSECOND: RelValueType
REL_VALUE_TYPE_MILLISECOND: RelValueType
REL_VALUE_TYPE_SECOND: RelValueType
REL_VALUE_TYPE_MINUTE: RelValueType
REL_VALUE_TYPE_HOUR: RelValueType
REL_VALUE_TYPE_DAY: RelValueType
REL_VALUE_TYPE_WEEK: RelValueType
REL_VALUE_TYPE_MONTH: RelValueType
REL_VALUE_TYPE_YEAR: RelValueType

class Declaration(_message.Message):
    __slots__ = ("loop",)
    DEF_FIELD_NUMBER: _ClassVar[int]
    LOOP_FIELD_NUMBER: _ClassVar[int]
    loop: Loop
    def __init__(self, loop: _Optional[_Union[Loop, _Mapping]] = ..., **kwargs) -> None: ...

class Def(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Loop(_message.Message):
    __slots__ = ("temporal_var", "inits", "body")
    TEMPORAL_VAR_FIELD_NUMBER: _ClassVar[int]
    INITS_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    temporal_var: LoopIndex
    inits: _containers.RepeatedCompositeFieldContainer[Def]
    body: _containers.RepeatedCompositeFieldContainer[Declaration]
    def __init__(self, temporal_var: _Optional[_Union[LoopIndex, _Mapping]] = ..., inits: _Optional[_Iterable[_Union[Def, _Mapping]]] = ..., body: _Optional[_Iterable[_Union[Declaration, _Mapping]]] = ...) -> None: ...

class Abstraction(_message.Message):
    __slots__ = ("vars", "value")
    VARS_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    vars: _containers.RepeatedCompositeFieldContainer[Var]
    value: Formula
    def __init__(self, vars: _Optional[_Iterable[_Union[Var, _Mapping]]] = ..., value: _Optional[_Union[Formula, _Mapping]] = ...) -> None: ...

class Formula(_message.Message):
    __slots__ = ("exists", "reduce", "conjunction", "disjunction", "ffi", "atom", "pragma", "primitive", "rel_atom", "cast")
    EXISTS_FIELD_NUMBER: _ClassVar[int]
    REDUCE_FIELD_NUMBER: _ClassVar[int]
    CONJUNCTION_FIELD_NUMBER: _ClassVar[int]
    DISJUNCTION_FIELD_NUMBER: _ClassVar[int]
    NOT_FIELD_NUMBER: _ClassVar[int]
    FFI_FIELD_NUMBER: _ClassVar[int]
    ATOM_FIELD_NUMBER: _ClassVar[int]
    PRAGMA_FIELD_NUMBER: _ClassVar[int]
    PRIMITIVE_FIELD_NUMBER: _ClassVar[int]
    REL_ATOM_FIELD_NUMBER: _ClassVar[int]
    CAST_FIELD_NUMBER: _ClassVar[int]
    exists: Exists
    reduce: Reduce
    conjunction: Conjunction
    disjunction: Disjunction
    ffi: FFI
    atom: Atom
    pragma: Pragma
    primitive: Primitive
    rel_atom: RelAtom
    cast: Cast
    def __init__(self, exists: _Optional[_Union[Exists, _Mapping]] = ..., reduce: _Optional[_Union[Reduce, _Mapping]] = ..., conjunction: _Optional[_Union[Conjunction, _Mapping]] = ..., disjunction: _Optional[_Union[Disjunction, _Mapping]] = ..., ffi: _Optional[_Union[FFI, _Mapping]] = ..., atom: _Optional[_Union[Atom, _Mapping]] = ..., pragma: _Optional[_Union[Pragma, _Mapping]] = ..., primitive: _Optional[_Union[Primitive, _Mapping]] = ..., rel_atom: _Optional[_Union[RelAtom, _Mapping]] = ..., cast: _Optional[_Union[Cast, _Mapping]] = ..., **kwargs) -> None: ...

class Exists(_message.Message):
    __slots__ = ("body",)
    BODY_FIELD_NUMBER: _ClassVar[int]
    body: Abstraction
    def __init__(self, body: _Optional[_Union[Abstraction, _Mapping]] = ...) -> None: ...

class Reduce(_message.Message):
    __slots__ = ("op", "body", "terms")
    OP_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    op: Abstraction
    body: Abstraction
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, op: _Optional[_Union[Abstraction, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Conjunction(_message.Message):
    __slots__ = ("args",)
    ARGS_FIELD_NUMBER: _ClassVar[int]
    args: _containers.RepeatedCompositeFieldContainer[Formula]
    def __init__(self, args: _Optional[_Iterable[_Union[Formula, _Mapping]]] = ...) -> None: ...

class Disjunction(_message.Message):
    __slots__ = ("args",)
    ARGS_FIELD_NUMBER: _ClassVar[int]
    args: _containers.RepeatedCompositeFieldContainer[Formula]
    def __init__(self, args: _Optional[_Iterable[_Union[Formula, _Mapping]]] = ...) -> None: ...

class Not(_message.Message):
    __slots__ = ("arg",)
    ARG_FIELD_NUMBER: _ClassVar[int]
    arg: Formula
    def __init__(self, arg: _Optional[_Union[Formula, _Mapping]] = ...) -> None: ...

class FFI(_message.Message):
    __slots__ = ("name", "args", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    args: _containers.RepeatedCompositeFieldContainer[Abstraction]
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[Abstraction, _Mapping]]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Atom(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Pragma(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Primitive(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[RelTerm]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[RelTerm, _Mapping]]] = ...) -> None: ...

class RelAtom(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[RelTerm]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[RelTerm, _Mapping]]] = ...) -> None: ...

class Cast(_message.Message):
    __slots__ = ("type", "input", "result")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    INPUT_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    type: RelType
    input: Term
    result: Term
    def __init__(self, type: _Optional[_Union[RelType, _Mapping]] = ..., input: _Optional[_Union[Term, _Mapping]] = ..., result: _Optional[_Union[Term, _Mapping]] = ...) -> None: ...

class RelTerm(_message.Message):
    __slots__ = ("specialized_value", "term")
    SPECIALIZED_VALUE_FIELD_NUMBER: _ClassVar[int]
    TERM_FIELD_NUMBER: _ClassVar[int]
    specialized_value: SpecializedValue
    term: Term
    def __init__(self, specialized_value: _Optional[_Union[SpecializedValue, _Mapping]] = ..., term: _Optional[_Union[Term, _Mapping]] = ...) -> None: ...

class Term(_message.Message):
    __slots__ = ("var", "constant")
    VAR_FIELD_NUMBER: _ClassVar[int]
    CONSTANT_FIELD_NUMBER: _ClassVar[int]
    var: Var
    constant: Constant
    def __init__(self, var: _Optional[_Union[Var, _Mapping]] = ..., constant: _Optional[_Union[Constant, _Mapping]] = ...) -> None: ...

class Var(_message.Message):
    __slots__ = ("name", "type")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: RelType
    def __init__(self, name: _Optional[str] = ..., type: _Optional[_Union[RelType, _Mapping]] = ...) -> None: ...

class Constant(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: PrimitiveValue
    def __init__(self, value: _Optional[_Union[PrimitiveValue, _Mapping]] = ...) -> None: ...

class SpecializedValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: PrimitiveValue
    def __init__(self, value: _Optional[_Union[PrimitiveValue, _Mapping]] = ...) -> None: ...

class Attribute(_message.Message):
    __slots__ = ("name", "args")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    name: str
    args: _containers.RepeatedCompositeFieldContainer[Constant]
    def __init__(self, name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[Constant, _Mapping]]] = ...) -> None: ...

class RelationId(_message.Message):
    __slots__ = ("id_low", "id_high")
    ID_LOW_FIELD_NUMBER: _ClassVar[int]
    ID_HIGH_FIELD_NUMBER: _ClassVar[int]
    id_low: int
    id_high: int
    def __init__(self, id_low: _Optional[int] = ..., id_high: _Optional[int] = ...) -> None: ...

class LoopIndex(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class RelType(_message.Message):
    __slots__ = ("primitive_type", "value_type")
    PRIMITIVE_TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_TYPE_FIELD_NUMBER: _ClassVar[int]
    primitive_type: PrimitiveType
    value_type: RelValueType
    def __init__(self, primitive_type: _Optional[_Union[PrimitiveType, str]] = ..., value_type: _Optional[_Union[RelValueType, str]] = ...) -> None: ...

class PrimitiveValue(_message.Message):
    __slots__ = ("string_value", "int_value", "float_value", "uint128_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    UINT128_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    int_value: int
    float_value: float
    uint128_value: UInt128
    def __init__(self, string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., float_value: _Optional[float] = ..., uint128_value: _Optional[_Union[UInt128, _Mapping]] = ...) -> None: ...

class UInt128(_message.Message):
    __slots__ = ("low", "high")
    LOW_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    low: int
    high: int
    def __init__(self, low: _Optional[int] = ..., high: _Optional[int] = ...) -> None: ...



import lqp.ir as ir
from lqp.proto.v1 import logic_pb2, fragments_pb2, transactions_pb2
from typing import Union, Dict, Any

def convert_primitive_type(pt: ir.PrimitiveType) -> logic_pb2.PrimitiveType:
    if pt == ir.PrimitiveType.STRING:
        return logic_pb2.PRIMITIVE_TYPE_STRING
    elif pt == ir.PrimitiveType.INT:
        return logic_pb2.PRIMITIVE_TYPE_INT
    elif pt == ir.PrimitiveType.FLOAT:
        return logic_pb2.PRIMITIVE_TYPE_FLOAT
    elif pt == ir.PrimitiveType.UINT128:
        return logic_pb2.PRIMITIVE_TYPE_UINT128
    else:
        return logic_pb2.PRIMITIVE_TYPE_UNSPECIFIED

def convert_rel_value_type(rvt: ir.RelValueType) -> logic_pb2.RelValueType:
    try:
        return getattr(logic_pb2.RelValueType, f"REL_VALUE_TYPE_{rvt.name}")
    except AttributeError:
        return logic_pb2.REL_VALUE_TYPE_UNSPECIFIED

def convert_rel_type(rt: ir.RelType) -> logic_pb2.RelType:
    if isinstance(rt, ir.PrimitiveType):
        return logic_pb2.RelType(primitive_type=convert_primitive_type(rt))
    elif isinstance(rt, ir.RelValueType):
        return logic_pb2.RelType(value_type=convert_rel_value_type(rt))
    else:
        # Default or error case
        return logic_pb2.RelType(primitive_type=logic_pb2.PRIMITIVE_TYPE_UNSPECIFIED)

def convert_uint128(val: ir.UInt128) -> logic_pb2.UInt128:
    return logic_pb2.UInt128(low=val.low, high=val.high)

def convert_primitive_value(pv: ir.PrimitiveValue) -> logic_pb2.PrimitiveValue:
    if isinstance(pv, str):
        return logic_pb2.PrimitiveValue(string_value=pv)
    elif isinstance(pv, int):
        return logic_pb2.PrimitiveValue(int_value=pv)
    elif isinstance(pv, float):
        return logic_pb2.PrimitiveValue(float_value=pv)
    elif isinstance(pv, ir.UInt128):
        return logic_pb2.PrimitiveValue(uint128_value=convert_uint128(pv))
    else:
        raise TypeError(f"Unsupported PrimitiveValue type: {type(pv)}")

def convert_constant(c: ir.Constant) -> logic_pb2.Constant:
    return logic_pb2.Constant(value=convert_primitive_value(c))

def convert_var(v: ir.Var) -> logic_pb2.Var:
    return logic_pb2.Var(name=v.name, type=convert_rel_type(v.type))

def convert_term(t: ir.Term) -> logic_pb2.Term:
    if isinstance(t, ir.Var):
        return logic_pb2.Term(var=convert_var(t))
    else:
        return logic_pb2.Term(constant=convert_constant(t))

def convert_relterm(t: ir.RelTerm) -> logic_pb2.RelTerm:
    if isinstance(t, ir.Specialized):
        val = logic_pb2.SpecializedValue(value=convert_primitive_value(t.value))
        return logic_pb2.RelTerm(specialized_value=val)
    else:
        return logic_pb2.RelTerm(term=convert_term(t))

def convert_relation_id(rid: ir.RelationId) -> logic_pb2.RelationId:
    return logic_pb2.RelationId(id_low=rid.id_low, id_high=rid.id_high)

def convert_fragment_id(fid: ir.FragmentId) -> fragments_pb2.FragmentId:
    return fragments_pb2.FragmentId(id=fid.id)

def convert_attribute(attr: ir.Attribute) -> logic_pb2.Attribute:
    return logic_pb2.Attribute(
        name=attr.name,
        args=[convert_constant(arg) for arg in attr.args]
    )

def convert_abstraction(abst: ir.Abstraction) -> logic_pb2.Abstraction:
    return logic_pb2.Abstraction(
        vars=[convert_var(v) for v in abst.vars],
        value=convert_formula(abst.value)
    )

def convert_formula(f: ir.Formula) -> logic_pb2.Formula:
    if isinstance(f, ir.Exists):
        return logic_pb2.Formula(exists=logic_pb2.Exists(body=convert_abstraction(f.body)))
    elif isinstance(f, ir.Reduce):
        return logic_pb2.Formula(reduce=logic_pb2.Reduce(
            op=convert_abstraction(f.op),
            body=convert_abstraction(f.body),
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Conjunction):
        return logic_pb2.Formula(conjunction=logic_pb2.Conjunction(args=[convert_formula(arg) for arg in f.args]))
    elif isinstance(f, ir.Disjunction):
        return logic_pb2.Formula(disjunction=logic_pb2.Disjunction(args=[convert_formula(arg) for arg in f.args]))
    elif isinstance(f, ir.Not):
        return logic_pb2.Formula(**{"not": logic_pb2.Not(arg=convert_formula(f.arg))}) # type: ignore
    elif isinstance(f, ir.FFI):
        return logic_pb2.Formula(ffi=logic_pb2.FFI(
            name=f.name,
            args=[convert_abstraction(arg) for arg in f.args],
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Atom):
        return logic_pb2.Formula(atom=logic_pb2.Atom(
            name=convert_relation_id(f.name),
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Pragma):
        return logic_pb2.Formula(pragma=logic_pb2.Pragma(
            name=f.name,
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Primitive):
        primitive_proto = logic_pb2.Primitive(name=f.name, terms=[convert_relterm(t) for t in f.terms])
        return logic_pb2.Formula(primitive=primitive_proto)
    elif isinstance(f, ir.RelAtom):
        rel_atom_proto = logic_pb2.RelAtom(name=f.name, terms=[convert_relterm(t) for t in f.terms])
        return logic_pb2.Formula(rel_atom=rel_atom_proto)
    elif isinstance(f, ir.Cast):
        return logic_pb2.Formula(cast=logic_pb2.Cast(
            type=convert_rel_type(f.type),
            input=convert_term(f.input),
            result=convert_term(f.result)
        ))
    else:
        raise TypeError(f"Unsupported Formula type: {type(f)}")

def convert_loop_index(li: ir.LoopIndex) -> logic_pb2.LoopIndex:
    return logic_pb2.LoopIndex(name=li.name)

def convert_def(d: ir.Def) -> logic_pb2.Def:
    return logic_pb2.Def(
        name=convert_relation_id(d.name),
        body=convert_abstraction(d.body),
        attrs=[convert_attribute(attr) for attr in d.attrs]
    )

def convert_loop(l: ir.Loop) -> logic_pb2.Loop:
    return logic_pb2.Loop(
        temporal_var=convert_loop_index(l.temporal_var),
        inits=[convert_def(init_def) for init_def in l.inits],
        body=[convert_declaration(decl) for decl in l.body]
    )

def convert_declaration(decl: ir.Declaration) -> logic_pb2.Declaration:
    if isinstance(decl, ir.Def):
        # 'def' is a Python keyword, so we use the same approach as with 'not'
        from typing import Dict, Any
        decl_dict: Dict[str, Any] = {'def': convert_def(decl)}
        return logic_pb2.Declaration(**decl_dict)  # type: ignore
    elif isinstance(decl, ir.Loop):
        return logic_pb2.Declaration(loop=convert_loop(decl))
    else:
        raise TypeError(f"Unsupported Declaration type: {type(decl)}")

def convert_fragment(frag: ir.Fragment) -> fragments_pb2.Fragment:
    return fragments_pb2.Fragment(
        id=convert_fragment_id(frag.id),
        declarations=[convert_declaration(decl) for decl in frag.declarations]
    )

def convert_define(d: ir.Define) -> transactions_pb2.Define:
    return transactions_pb2.Define(fragment=convert_fragment(d.fragment))

def convert_undefine(u: ir.Undefine) -> transactions_pb2.Undefine:
    return transactions_pb2.Undefine(fragment_id=convert_fragment_id(u.fragment_id))

def convert_context(c: ir.Context) -> transactions_pb2.Context:
    return transactions_pb2.Context(relations=[convert_relation_id(rid) for rid in c.relations])

def convert_write(w: ir.Write) -> transactions_pb2.Write:
    wt = w.write_type
    if isinstance(wt, ir.Define):
        return transactions_pb2.Write(define=convert_define(wt))
    elif isinstance(wt, ir.Undefine):
        return transactions_pb2.Write(undefine=convert_undefine(wt))
    elif isinstance(wt, ir.Context):
        return transactions_pb2.Write(context=convert_context(wt))
    else:
        raise TypeError(f"Unsupported Write type: {type(wt)}")

def convert_demand(d: ir.Demand) -> transactions_pb2.Demand:
    return transactions_pb2.Demand(relation_id=convert_relation_id(d.relation_id))

def convert_output(o: ir.Output) -> transactions_pb2.Output:
    kwargs: Dict[str, Any] = {'relation_id': convert_relation_id(o.relation_id)}
    if o.name is not None:
        kwargs['name'] = o.name
    return transactions_pb2.Output(**kwargs) # type: ignore

def convert_abort(a: ir.Abort) -> transactions_pb2.Abort:
    kwargs: Dict[str, Any] = {'relation_id': convert_relation_id(a.relation_id)}
    if a.name is not None:
        kwargs['name'] = a.name
    return transactions_pb2.Abort(**kwargs) # type: ignore

def convert_whatif(wi: ir.WhatIf) -> transactions_pb2.WhatIf:
    kwargs: Dict[str, Any] = {'epoch': convert_epoch(wi.epoch)}
    if wi.branch is not None:
        kwargs['branch'] = wi.branch
    return transactions_pb2.WhatIf(**kwargs) # type: ignore

def convert_read(r: ir.Read) -> transactions_pb2.Read:
    rt = r.read_type
    if isinstance(rt, ir.Demand):
        return transactions_pb2.Read(demand=convert_demand(rt))
    elif isinstance(rt, ir.Output):
        return transactions_pb2.Read(output=convert_output(rt))
    elif isinstance(rt, ir.WhatIf):
        return transactions_pb2.Read(what_if=convert_whatif(rt)) # Note the underscore
    elif isinstance(rt, ir.Abort):
        return transactions_pb2.Read(abort=convert_abort(rt))
    else:
        raise TypeError(f"Unsupported Read type: {type(rt)}")

def convert_epoch(e: ir.Epoch) -> transactions_pb2.Epoch:
    return transactions_pb2.Epoch(
        persistent_writes=[convert_write(w) for w in e.persistent_writes],
        local_writes=[convert_write(w) for w in e.local_writes],
        reads=[convert_read(r) for r in e.reads]
    )

def convert_transaction(t: ir.Transaction) -> transactions_pb2.Transaction:
    return transactions_pb2.Transaction(
        epochs=[convert_epoch(e) for e in t.epochs]
    )

def ir_to_proto(node: ir.LqpNode) -> Union[
    transactions_pb2.Transaction,
    fragments_pb2.Fragment,
    logic_pb2.Declaration,
    logic_pb2.Formula,
    logic_pb2.Term
]:
    if isinstance(node, ir.Transaction):
        return convert_transaction(node)
    elif isinstance(node, ir.Fragment):
        return convert_fragment(node)
    elif isinstance(node, ir.Declaration):
        return convert_declaration(node)
    elif isinstance(node, ir.Formula):
        return convert_formula(node)
    elif isinstance(node, ir.Term):
        return convert_term(node)
    else:
        raise TypeError(f"Unsupported top-level IR node type for conversion: {type(node)}")



import argparse
import os
import hashlib
from lark import Lark, Transformer
from typing import cast
import lqp.ir as ir
from lqp.emit import ir_to_proto
from lqp.validator import validate_lqp, fill_types
from google.protobuf.json_format import MessageToJson

grammar = """
start: transaction | fragment

transaction: "(transaction" epoch* ")"
epoch: "(epoch" persistent_writes? local_writes? reads? ")"
persistent_writes: "(persistent_writes" write* ")"
local_writes: "(local_writes" write* ")"
reads: "(reads" read* ")"

write: define | undefine | context
define: "(define" fragment ")"
undefine: "(undefine" fragment_id ")"
context: "(context" relation_id* ")"

read: demand | output | abort
demand: "(demand" relation_id ")"
output: "(output" name? relation_id ")"
abort: "(abort" name? relation_id ")"

fragment: "(fragment" fragment_id declaration* ")"

declaration: def_
def_: "(def" relation_id abstraction attrs? ")"

abstraction: "(" vars formula ")"
vars: "[" var* "]"

formula: exists | reduce | conjunction | disjunction | not_ | ffi | atom | pragma | primitive | true | false | relatom | cast
exists: "(exists" vars formula ")"
reduce: "(reduce" abstraction abstraction terms ")"
conjunction: "(and" formula* ")"
disjunction: "(or" formula* ")"
not_: "(not" formula ")"
ffi: "(ffi" name args terms ")"
atom: "(atom" relation_id term* ")"
relatom: "(relatom" name relterm* ")"
cast: "(cast" rel_type term term ")"
pragma: "(pragma" name terms ")"
true: "(true)"
false: "(false)"

args: "(args" abstraction* ")"
terms: "(terms" term* ")"

primitive: raw_primitive | eq | lt | lt_eq | gt | gt_eq | add | minus | multiply | divide
raw_primitive: "(primitive" name relterm* ")"
eq: "(=" term term ")"
lt: "(<" term term ")"
lt_eq: "(<=" term term ")"
gt: "(>" term term ")"
gt_eq: "(>=" term term ")"

add: "(+" term term term ")"
minus: "(-" term term term ")"
multiply: "(*" term term term ")"
divide: "(/" term term term ")"

relterm: specialized_value | term
term: var | constant
var: SYMBOL "::" rel_type | SYMBOL
constant: primitive_value

attrs: "(attrs" attribute* ")"
attribute: "(attribute" name constant* ")"

fragment_id: ":" SYMBOL
relation_id: (":" SYMBOL) | NUMBER
name: ":" SYMBOL

specialized_value: "#" primitive_value

primitive_value: STRING | NUMBER | FLOAT | UINT128

rel_type: PRIMITIVE_TYPE | REL_VALUE_TYPE
PRIMITIVE_TYPE: "STRING" | "INT" | "FLOAT" | "UINT128" | "ENTITY"
REL_VALUE_TYPE: "DECIMAL" | "DATE" | "DATETIME"
              | "NANOSECOND" | "MICROSECOND" | "MILLISECOND" | "SECOND" | "MINUTE" | "HOUR"
              | "DAY" | "WEEK" | "MONTH" | "YEAR"

SYMBOL: /[a-zA-Z_][a-zA-Z0-9_-]*/
STRING: "\\"" /[^"]*/ "\\""
NUMBER: /\\d+/
UINT128: /0x[0-9a-fA-F]+/
FLOAT: /\\d+\\.\\d+/

COMMENT: /;;.*/  // Matches ;; followed by any characters except newline
%ignore /\\s+/
%ignore COMMENT
"""


def desugar_to_raw_primitive(name, terms):
    # Convert terms to relterms
    return ir.Primitive(name=name, terms=terms)

class LQPTransformer(Transformer):
    def start(self, items):
        return items[0]

    def PRIMITIVE_TYPE(self, s):
        # Map ENTITY -> HASH
        if s.upper() == "ENTITY":
            return ir.PrimitiveType.UINT128
        return getattr(ir.PrimitiveType, s.upper())

    def REL_VALUE_TYPE(self, s):
        return getattr(ir.RelValueType, s.upper())

    def rel_type(self, items):
        return items[0]

    #
    # Transactions
    #
    def transaction(self, items):
        return ir.Transaction(epochs=items)
    def epoch(self, items):
        kwargs = {k: v for k, v in items if v} # Filter out None values
        return ir.Epoch(**kwargs)

    def persistent_writes(self, items):
        return ("persistent_writes", items)
    def local_writes(self, items):
        return ("local_writes", items)
    def reads(self, items):
        return ("reads", items)
    def write(self, items):
        return ir.Write(write_type=items[0])

    def define(self, items):
        return ir.Define(fragment=items[0])

    def undefine(self, items):
        return ir.Undefine(fragment_id=items[0])

    def context(self, items):
        return ir.Context(relations=items)

    def read(self, items):
        return ir.Read(read_type=items[0])
    def demand(self, items):
        return ir.Demand(relation_id=items[0])

    def output(self, items):
        if len(items) == 1:
            return ir.Output(name=None, relation_id=items[0])
        return ir.Output(name=items[0], relation_id=items[1])

    def abort(self, items):
        if len(items) == 1:
            return ir.Abort(name=None, relation_id=items[0])
        return ir.Abort(name=items[0], relation_id=items[1])

    #
    # Logic
    #
    def fragment(self, items):
        return ir.Fragment(id=items[0], declarations=items[1:])

    def fragment_id(self, items):
        return ir.FragmentId(id=items[0].encode())

    def declaration(self, items):
        return items[0]
    def def_(self, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Def(name=name, body=body, attrs=attrs)

    def abstraction(self, items):
        return ir.Abstraction(vars=items[0], value=items[1])

    def vars(self, items):
        return items
    def attrs(self, items):
        return items

    def formula(self, items):
        return items[0]
    def true(self, _):
        return ir.Conjunction(args=[])

    def false(self, _):
        return ir.Disjunction(args=[])

    def exists(self, items):
        # Create Abstraction for body directly here
        body_abstraction = ir.Abstraction(vars=items[0], value=items[1])
        return ir.Exists(body=body_abstraction)

    def reduce(self, items):
        return ir.Reduce(op=items[0], body=items[1], terms=items[2])

    def conjunction(self, items):
        return ir.Conjunction(args=items)

    def disjunction(self, items):
        return ir.Disjunction(args=items)

    def not_(self, items):
        return ir.Not(arg=items[0])

    def ffi(self, items):
        return ir.FFI(name=items[0], args=items[1], terms=items[2])

    def atom(self, items):
        return ir.Atom(name=items[0], terms=items[1:])

    def pragma(self, items):
        return ir.Pragma(name=items[0], terms=items[1])

    def relatom(self, items):
        return ir.RelAtom(name=items[0], terms=items[1:])

    def cast(self, items):
        return ir.Cast(type=items[0], input=items[1], result=items[2])

    #
    # Primitives
    #
    def primitive(self, items):
        if isinstance(items[0], ir.Formula):
            return items[0]
        raise TypeError(f"Unexpected primitive type: {type(items[0])}")
    def raw_primitive(self, items):
        return ir.Primitive(name=items[0], terms=items[1:])
    def _make_primitive(self, name_symbol, terms):
         # Convert name symbol to string if needed, assuming self.name handles it
         name_str = self.name([name_symbol]) if isinstance(name_symbol, str) else name_symbol
         return ir.Primitive(name=name_str, terms=terms)
    def eq(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_eq"]), items)
    def lt(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_lt_monotype"]), items)
    def lt_eq(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_lt_eq_monotype"]), items)
    def gt(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_gt_monotype"]), items)
    def gt_eq(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_gt_eq_monotype"]), items)

    def add(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_add_monotype"]), items)
    def minus(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_subtract_monotype"]), items)
    def multiply(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_multiply_monotype"]), items)
    def divide(self, items):
        return desugar_to_raw_primitive(self.name(["rel_primitive_divide_monotype"]), items)

    def args(self, items):
        return items
    def terms(self, items):
        return items

    def relterm(self, items):
        return items[0]
    def term(self, items):
        return items[0]
    def var(self, items):
        identifier = items[0]
        if len(items) > 1:
            rel_type_obj = items[1]
            return ir.Var(name=identifier, type=rel_type_obj)
        else:
            return ir.Var(name=identifier, type=ir.PrimitiveType.UNSPECIFIED)
    def constant(self, items):
        return items[0]
    def specialized_value(self, items):
        return ir.Specialized(value=items[0])

    def name(self, items):
        return items[0] # SYMBOL string

    def attribute(self, items):
        return ir.Attribute(name=items[0], args=items[1:])

    def relation_id(self, items):
        ident = items[0] # Remove leading ':'
        if isinstance(ident, str):
            hash_val = int(hashlib.sha256(ident.encode()).hexdigest()[:16], 16) # First 64 bits of SHA-256
            return ir.RelationId(id_low=hash_val, id_high=0) # Simplified hashing
        elif isinstance(ident, int):
            low = ident & 0xFFFFFFFFFFFFFFFF
            high = (ident >> 64) & 0xFFFFFFFFFFFFFFFF
            return ir.RelationId(id_low=low, id_high=high)

    #
    # Primitive values
    #
    def primitive_value(self, items):
        return items[0]
    def STRING(self, s):
        return s[1:-1] # Strip quotes
    def NUMBER(self, n):
        return int(n)
    def FLOAT(self, f):
        return float(f)
    def SYMBOL(self, sym):
        return str(sym)
    def UINT128(self, u):
        uint128_val = int(u, 16)
        low = uint128_val & 0xFFFFFFFFFFFFFFFF
        high = (uint128_val >> 64) & 0xFFFFFFFFFFFFFFFF
        return ir.UInt128(low=low, high=high)

# LALR(1) is significantly faster than Earley for parsing, especially on larger inputs. It
# uses a precomputed parse table, reducing runtime complexity to O(n) (linear in input
# size), whereas Earley is O(n³) in the worst case (though often O(n²) or better for
# practical grammars). The LQP grammar is relatively complex but unambiguous, making
# LALR(1)’s speed advantage appealing for a CLI tool where quick parsing matters.
parser = Lark(grammar, parser="lalr", transformer=LQPTransformer())

def parse_lqp(text) -> ir.LqpNode:
    """Parse LQP text and return an IR node that can be converted to protocol buffers"""
    lqp_node = cast(ir.LqpNode, parser.parse(text))
    fill_types(lqp_node)
    return lqp_node

def process_file(filename, bin, json):
    with open(filename, "r") as f:
        lqp_text = f.read()

    lqp = parse_lqp(lqp_text)
    validate_lqp(lqp)
    lqp_proto = ir_to_proto(lqp)
    print(lqp_proto)

    # Write binary output to the configured directories, using the same filename.
    if bin:
        with open(bin, "wb") as f:
            f.write(lqp_proto.SerializeToString())

    # Write JSON output
    if json:
        with open(json, "w") as f:
            f.write(MessageToJson(lqp_proto, preserving_proto_field_name=True))

def main():
    arg_parser = argparse.ArgumentParser(description="Parse LQP S-expression into Protobuf binary and JSON files.")
    arg_parser.add_argument("input_directory", help="path to the input LQP S-expression files")
    arg_parser.add_argument("--bin", help="output directory for the binary encoded protobuf")
    arg_parser.add_argument("--json", help="output directory for the JSON encoded protobuf")
    args = arg_parser.parse_args()

    print(args)

    # Check if directory
    if not os.path.isdir(args.input_directory):
        filename = args.input_directory
        if not filename.endswith(".lqp"):
            print(f"Skipping file {filename} as it does not have the .lqp extension")
            return

        bin = args.bin if args.bin else None
        if bin and not bin.endswith(".bin"):
            print(f"Skipping output {bin} as it does not have the .bin extension")
            bin = None

        json = args.json if args.json else None
        if json and not json.endswith(".json"):
            print(f"Skipping output {json} as it does not have the .json extension")
            json = None
        process_file(filename, bin, json)

    else:
        # Process each file in the input directory
        for file in os.listdir(args.input_directory):
            if not file.endswith(".lqp"):
                print(f"Skipping file {file} as it does not have the .lqp extension")
                continue

            filename = os.path.join(args.input_directory, file)
            basename = os.path.splitext(file)[0]
            bin = os.path.join(args.bin, basename+".bin") if args.bin else None
            json = os.path.join(args.json, basename+".json") if args.json else None
            process_file(filename, bin, json)


if __name__ == "__main__":
    main()
